package thesis.jndi.provider.impl;

import javax.naming.spi.NamingManager;
import javax.naming.spi.StateFactory;
import javax.naming.spi.ObjectFactory;
import javax.naming.spi.InitialContextFactory;
import javax.naming.NamingException;
import javax.naming.Name;
import javax.naming.Context;
import javax.naming.NoInitialContextException;

import thesis.jndi.provider.api.DispatchStateFactory;
import thesis.jndi.provider.api.Provider;
import thesis.jndi.provider.api.ProviderManager;
import thesis.jndi.provider.impl.context.InitialContextFactoryBuilderImpl;
import thesis.jndi.provider.impl.semantics.ObjectFactoryBuilderImpl;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;

/* Full implementation of ProviderManager
 * 
 * @author Elitsa Pancheva
 */
public class ProviderResolver extends ProviderManager {

  /**
   *  boolean flag showing whether an instance of ProviderResolver has been already created or not
   */
  private static boolean initialized    = false;

  /**
   *  all the registered Provider implementations
   */
  private Hashtable providers = new Hashtable();
   /**
   *  keeps the class name of the last successful Provider for a given initial context factory name
   */
  private Hashtable icfProviderMapping = new Hashtable();
  /**
   *  keeps the class name of the last successful Provider for a given object factory name
   */
  private Hashtable ofProviderMapping = new Hashtable();
  /**
   *  keeps the class name of the last successful Provider for a given state factory name
   */
  private Hashtable sfProviderMapping = new Hashtable();
  /**
   *  keeps the class name of the last successful Provider for a given url context factory name
   */
  private Hashtable urlcfProviderMapping = new Hashtable();

  static {
    if (!NamingManager.hasInitialContextFactoryBuilder()) {
      try {
        instance = new ProviderResolver();
        NamingManager.setInitialContextFactoryBuilder(new InitialContextFactoryBuilderImpl());
        NamingManager.setObjectFactoryBuilder(new ObjectFactoryBuilderImpl());
        System.out.println("[ProviderResolver] InitialContextFactoryBuilder is set in NamingManager");
        System.out.println("[ProviderResolver] ObjectFactoryBuilder is set in NamingManager");
      } catch (NamingException e) {
        System.out.println("[ProviderResolver] failed to set InitialContext/Object Factory in javax.naming.spi.NamingManager because of: " + e.toString());
      }
      initialized = true;
      System.out.println("[ProviderResolver] initialized successfully");
    }
  }

  public static synchronized void init() throws Exception {
    if (!initialized) {
      instance = new ProviderResolver();
      if (!NamingManager.hasInitialContextFactoryBuilder()) {
        NamingManager.setInitialContextFactoryBuilder(new InitialContextFactoryBuilderImpl());
        NamingManager.setObjectFactoryBuilder(new ObjectFactoryBuilderImpl());
        System.out.println("[ProviderResolver] InitialContextFactoryBuilder is set in NamingManager");
        System.out.println("[ProviderResolver] ObjectFactoryBuilder is set in NamingManager");
      }
      initialized = true;
      System.out.println("[ProviderResolver] initialized successfully");
    }
  }

  public void registerProvider(Provider provider) throws NamingException {
    System.out.println("Provide: " + provider + " is going to be registered in the ProviderResolver");
    System.out.println("Before the registration the hashtable of providers contains: " + providers);
    
    synchronized (this.providers) {
      if (provider != null) {
    	providers.put(provider.getClass().getName(), provider);
      }
    }
    System.out.println("After registration the hashtable of providers contains: " + providers);
  }

  public void unregisterProvider(Provider provider) throws NamingException {
    System.out.println("Provider: " + provider + " is going to be unregistered from the ProviderResolver");
    System.out.println("Before the unregistration the hashtable of providers contains: " + providers);
   
    if (provider != null) {
      Object r = null;

      synchronized (this.providers) {
        r = providers.remove(provider.getClass().getName());
      }

      if (r != null) {
        Enumeration keys = null;
        Object key = null;
        Object value = null;

        synchronized (icfProviderMapping) {
          for (keys = icfProviderMapping.keys(); keys.hasMoreElements(); ) {
            key = keys.nextElement();
            value = icfProviderMapping.get(key);
            if (value.equals(provider.getClass().getName())) {
            	icfProviderMapping.remove(key);
            }
          }
        }

        synchronized (ofProviderMapping) {
          for (keys = ofProviderMapping.keys(); keys.hasMoreElements(); ) {
            key = keys.nextElement();
            value = ofProviderMapping.get(key);
            if (value.equals(provider.getClass().getName())) {
            	ofProviderMapping.remove(key);
            }
          }
        }

        synchronized (urlcfProviderMapping) {
          for (keys = urlcfProviderMapping.keys(); keys.hasMoreElements(); ) {
            key = keys.nextElement();
            value = urlcfProviderMapping.get(key);
            if (value.equals(provider.getClass().getName())) {
            	urlcfProviderMapping.remove(key);
            }
          }
        }

        synchronized (sfProviderMapping) {
          for (keys = sfProviderMapping.keys(); keys.hasMoreElements(); ) {
            key = keys.nextElement();
            value = sfProviderMapping.get(key);
            if (value.equals(provider.getClass().getName())) {
            	sfProviderMapping.remove(key);
            }
          }
        }
      }

      System.out.println("After unregistration the hashtable of providers contains: " + providers);
    }
  }

  /**
   * dispatch to state factories call..
   * @see javax.naming.spi.StateFactory#getStateToBind(java.lang.Object, javax.naming.Name, javax.naming.Context, java.util.Hashtable)
   */
  public Object getStateToBind(Object obj, Name name, Context nameCtx, Hashtable environment) throws NamingException {
    String efs = (String) environment.get(DispatchStateFactory.EXTRA_STATE_FACTORIES);
    String fs = getStateFactoriesFromEnvironemnt((String) environment.get(Context.STATE_FACTORIES));
    
    if (efs == null) {
      if (fs == null) {
        // no state factory => nothing to do.
        // it's better to return null => NaminManager will be able to continue with the default behavior
        return null;
      } else {
    	// no extra state factories specified => use only the ones defined in the standard property
    	efs = fs;  
      }
    } else { // we have extra state factories
      if (fs != null) { // if we have factories in both properties combine them
    	efs.concat(":"+fs);
      }	
    }
    StringTokenizer tokenizer = new StringTokenizer(efs, ":");
    StateFactory fac;
    Object result = null;
    while (tokenizer.hasMoreTokens()) {
      String nextToken = tokenizer.nextToken();
      fac = findStateFactory(nextToken);
      if (fac != null) {
        result = fac.getStateToBind(obj, name, nameCtx, environment);
        if (result!=null) {
          return result;
        }
      }
      // ok, we couldn't do it ourselves.
      // we still try to go by
      // the default JNDI behavior: load class (as a fall-back)
      // NOTE: unlike for object factories we wouldn't really need to do this..., since we do not
      // overload the default environment property
      try {
        Class factoryClass = Class.forName(nextToken, true, Thread.currentThread().getContextClassLoader());
        StateFactory loadedFactory = (StateFactory) factoryClass.newInstance();
        if (loadedFactory != null) {
          result = loadedFactory.getStateToBind(obj, name, nameCtx, environment);
          if (result != null) {
            return result;
          }
        }
        
     //   System.out.println("getStateToBind method of ProviderResolver was called, none of the resolvers was able to return a factory with name: " + nextToken + " so a new instance: " + loadedFactory + " was created using the context class loader : " + Thread.currentThread().getContextClassLoader());
        return loadedFactory;
      } catch (ClassNotFoundException cnfe) {
        System.out.println("ProviderResolver was not able to return an instance of state factory with name: " + nextToken);
        cnfe.printStackTrace();
      } catch (NoClassDefFoundError ncdfe) {
        System.out.println("ProviderResolver was not able to return an instance of state factory with name: " + nextToken);
        ncdfe.printStackTrace();
      } catch (Exception e) {
        System.out.println("ProviderResolver was not able to return an instance of state factory with name: " + nextToken);
        e.printStackTrace();
      }
    }
    // if no state factory available - return null so that NamingManager will continue with its default behaviour
    return null;
  }

  private String getStateFactoriesFromEnvironemnt(String wholeStr) {
  	if (wholeStr == null) {
  		return "";
  	}
		String toExclude = DispatchStateFactory.class.getName();
		if (wholeStr.startsWith(toExclude)) {
		  if (wholeStr.equals(toExclude)) {
			return null;
		  }
		  return wholeStr.substring(toExclude.length() + 1);
		} else if (wholeStr.endsWith(toExclude)) {
			return wholeStr.substring(0, wholeStr.length() - toExclude.length() -1);
		} else {
			return wholeStr.replaceFirst(toExclude + ":", "");
		}
  }

/**
   * finds a state factory for the given data with the "infrastructure"
   *
   * @param stateFactoryName
   * @return
   * @throws javax.naming.NamingException
   */
  public StateFactory findStateFactory(final String stateFactoryName) throws NamingException {
    Provider provider = null;
    StateFactory stateFactory = null;

    // try first to use the last Provider used to return this factory name
    Object providerClassName = null;
    providerClassName = sfProviderMapping.get(stateFactoryName);

    if (providerClassName != null) {
      provider = (Provider) providers.get(providerClassName);
    }

    if (provider != null) {
      try {
        stateFactory = provider.getStateFactory(stateFactoryName);
        if (stateFactory != null) {
     //     System.out.println("Last used provider: " + provider + " will return the state factory: " + stateFactory);
          return stateFactory;
        }
      } catch (Exception e) {
        NamingException ne = new NamingException("Failed to retrieve state factory " + stateFactoryName + ". Exception is: " + e.toString());
        ne.setRootCause(e);
        throw ne;
      }
    }

    // the last used Provider either is null or it can no longer load this factory => go through all of the providers
    synchronized (this.providers) {
      for (Enumeration elements  = providers.elements(); elements.hasMoreElements(); ) {
        provider = (Provider) elements.nextElement();
        try {
          stateFactory = provider.getStateFactory(stateFactoryName);
          if (stateFactory != null) {
        	sfProviderMapping.put(stateFactoryName,  provider.getClass().getName());
          //  System.out.println("Provider: " + provider + " will return the state factory: " + stateFactory);
            return stateFactory;
          }
        } catch (Exception e) {
          NamingException ne = new NamingException("Failed to retrieve state factory " + stateFactoryName + ". Exception is: " + e.toString());
          ne.setRootCause(e);
          throw ne;
        }
      }
    }
    // return null
    return stateFactory;
  }

  /**
   * finds an object factory for the given data with the "infrastructure"
   *
   * @param objectFactoryName
   * @return
   * @throws javax.naming.NamingException
   */
  public ObjectFactory findObjectFactory(final String objectFactoryName) throws NamingException, IllegalAccessException, InstantiationException {
    Provider provider = null;
    ObjectFactory objectFactory = null;
    Object providerClassName = null;

    // try first to use the last Provider used to return this factory name
    providerClassName = ofProviderMapping.get(objectFactoryName);

    if (providerClassName != null) {
      provider = (Provider) providers.get(providerClassName);
    }

    if (provider != null) {
      try {
        objectFactory = provider.getObjectFactory(objectFactoryName);
        if (objectFactory != null) {
        ///  System.out.println("Last used provider: " + provider + " will return the object factory: " + objectFactory);
          return objectFactory;
        }
      } catch (Exception e) {
        NamingException ne = new NamingException("Failed to retrieve object factory " + objectFactoryName + ". Exception is: " + e.toString());
        ne.setRootCause(e);
        throw ne;
      }
    }

    // the last used Provider either is null or it can no longer load this factory => go through all of the providers
    synchronized (this.providers) {
      for (Enumeration elements  = providers.elements(); elements.hasMoreElements(); ) {
        provider = (Provider) elements.nextElement();
        try {
          objectFactory = provider.getObjectFactory(objectFactoryName);
          if (objectFactory != null) {
        	ofProviderMapping.put(objectFactoryName,  provider.getClass().getName());
         //   System.out.println("Provider: " + provider + " will return the object factory: " + objectFactory);
            return objectFactory;
          }
        } catch (Exception e) {
          // this provider indicates that this factory can be resolved by it but an Exception occured
          NamingException ne = new NamingException("Failed to retrieve object factory " + objectFactoryName + ". Exception is: " + e.toString());
          ne.setRootCause(e);
          throw ne;
        }
      }
    }

    // ok, we couldn't do it ourselves.
    // this means we should still try to go by the default JNDI behavior:
    try {
      Class factoryClass = Class.forName(objectFactoryName, true, Thread.currentThread().getContextClassLoader());
      ObjectFactory loadedFactory = (ObjectFactory) factoryClass.newInstance();
     // System.out.println("findObjectFactory method of ProviderResolver was called, none of the providers was able to return a factory with name: " + objectFactoryName + " so a new instance: " + loadedFactory + " was created using the context class loader : " + Thread.currentThread().getContextClassLoader());
      return loadedFactory;
    } catch (ClassNotFoundException cnfe) {
      System.out.println("ProviderResolver was not able to return an instance of object factory with name: " + objectFactoryName);
      // its ok, we were just not able to provide it
      return null;
    } catch (NoClassDefFoundError ncdfe) {
      System.out.println("ProviderResolver was not able to return an instance of object factory with name: " + objectFactoryName);
      ncdfe.printStackTrace();
      // its ok, we were just not able to provide it
      return null;
    }
  }

  /**
   * finds an initial context factory for the given data with the "infrastructure"
   * @param env
   * @return Initial Context Factory implementation or <code>null</code> if none found
   * @throws javax.naming.NamingException in case of errors
   */
  public InitialContextFactory findInitialContextFactory(Hashtable env) throws NamingException {
	Provider provider = null;
    InitialContextFactory initCtxFactory = null;
    String factoryName = ((String) env.get(Context.INITIAL_CONTEXT_FACTORY));
    Object providerClassName = null;

    // try first to use the last Provider used to return this factory name
    if (factoryName != null) {
      providerClassName = icfProviderMapping.get(factoryName);

      if (providerClassName != null) {
        provider = (Provider) providers.get(providerClassName);
      }

      if (provider != null) {
        try {
          initCtxFactory = provider.getInitialContextFactory(env);
          if (initCtxFactory != null) {
         //   System.out.println("Last used provider: " + provider + " will return the initial context factory: " + initCtxFactory);
            return initCtxFactory;
          }
        } catch (Exception e) {
          NamingException ne = new NamingException("Failed to retrieve initial context factory " + factoryName + ". Exception is: " + e.toString());
          ne.setRootCause(e);
          throw ne;
        }
      }
    }

    // the last used Provider either is null or it can no longer load this factory => go through all of the providers
    synchronized (this.providers) {
//	      System.out.println("number of providers = " + providers.size());
      for (Enumeration elements  = providers.elements(); elements.hasMoreElements(); ) {
        provider = (Provider) elements.nextElement();
        try {
          initCtxFactory = provider.getInitialContextFactory(env);
          if (initCtxFactory != null) {
            if (factoryName != null) {
            	icfProviderMapping.put(factoryName,  provider.getClass().getName());
            }
         //   System.out.println("Provider: " + provider + " will return the initial context factory: " + initCtxFactory);
            return initCtxFactory;
          }
        } catch (Exception e) {
          NamingException ne = new NamingException("Failed to retrieve initial context factory " + env.get(Context.INITIAL_CONTEXT_FACTORY) + ". Exception is: " + e.toString());
          ne.setRootCause(e);
          throw ne;
        }
      }
    }

     try {
      if (factoryName != null) {
        Class factoryClass = Class.forName(factoryName, true, Thread.currentThread().getContextClassLoader());
        InitialContextFactory loadedFactory = (InitialContextFactory) factoryClass.newInstance();
     //   System.out.println("findInitialContextFactory method of ProviderResolver was called, none of the providers was able to return a factory with name: " + factoryName + " so a new instance: " + loadedFactory + " was created using the context class loader : " + Thread.currentThread().getContextClassLoader());
        return loadedFactory;
      }
    } catch (ClassNotFoundException cnfe) {
      System.out.println("ProviderResolver was not able to return an instance of initial context factory with name: " + factoryName);
      cnfe.printStackTrace();
      // its ok, we were just not able to provide it
      NoInitialContextException ne = new NoInitialContextException("Cannot instantiate class: " + factoryName);
      ne.setRootCause(cnfe);
      throw ne;
    } catch (NoClassDefFoundError ncdfe) {
      System.out.println("ProviderResolver was not able to return an instance of initial context factory with name: " + factoryName);
      ncdfe.printStackTrace();
      // its ok, we were just not able to provide it
      NoInitialContextException ne = new NoInitialContextException("Cannot instantiate class: " + factoryName);
      ne.setRootCause(ncdfe);
      throw ne;
    } catch (Exception e) {
      System.out.println("ProviderResolver was not able to return an instance of initial context factory with name: " + factoryName);
      NoInitialContextException ne = new NoInitialContextException("Cannot instantiate class: " + factoryName);
      ne.setRootCause(e);
      throw ne;
    }
    // return null
    return initCtxFactory;
  }
  /**
   * finds a url context factory for the given data with the "infrastructure"
   * @param scheme
   * @param env
   * @return
   * @throws javax.naming.NamingException
   */
  public ObjectFactory findURLContextFactory(final String scheme, Hashtable env) throws NamingException {
	Provider provider = null;
    ObjectFactory urlContextFactory = null;
    Object providerClassName = null;

    // try first to use the last Provider used to return this factory name
    providerClassName = urlcfProviderMapping.get(scheme);

    if (providerClassName != null) {
      provider = (Provider) providers.get(providerClassName);
    }

    if (provider != null) {
      try {
        urlContextFactory = provider.getURLContextFactory(scheme, env);
        if (urlContextFactory != null) {
   //       System.out.println("Last used provider: " + provider + " will return the URL context factory: " + urlContextFactory);
          return urlContextFactory;
        }
      } catch (Exception e) {
        NamingException ne = new NamingException("Failed to retrieve URL Context Factory for scheme " + scheme + ". Exception is: " + e.toString());
        ne.setRootCause(e);
        throw ne;
      }
    }

    // the last used Provider either is null or it can no longer load this factory => go through all of the providers
    synchronized (this.providers) {
      for (Enumeration elements  = providers.elements(); elements.hasMoreElements(); ) {
        provider = (Provider) elements.nextElement();
        try {
          urlContextFactory = provider.getURLContextFactory(scheme, env);
          if (urlContextFactory != null) {
        	urlcfProviderMapping.put(scheme, provider.getClass().getName());
      //      System.out.println("Provider: " + provider + " will return the URL context factory: " + urlContextFactory);
            return urlContextFactory;
          }
        } catch (Exception e) {
          NamingException ne = new NamingException("Failed to retrieve URL Context Factory for scheme : " + scheme + ". Exception is: " + e.toString());
          ne.setRootCause(e);
          throw ne;
          //continue with the next provider
        }
      }
    }
    // return null
    return urlContextFactory;
  }

  
}


