package thesis.jndi.provider.impl.context;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.naming.Name;
import javax.naming.NoInitialContextException;
import javax.naming.spi.ObjectFactory;
import javax.naming.spi.NamingManager;
import javax.naming.spi.InitialContextFactory;

import thesis.jndi.provider.impl.ProviderResolver;

import java.util.Hashtable;

/* Default InitialContext implementation. This class overrides the default behaviour of InitilaContext.getURLOrDefaultInitCtx(..)
 * in order to involve ProviderResolver in resolving URLContextFactory implementation.
 *
 * @author Elitsa Pancheva
 */
public class DefaultInitialContext extends InitialContext {
	
  private Context defaultInitContext = null;
  /**
   * Constructor
   */ 
  public DefaultInitialContext(Hashtable environment) throws NamingException {
    super(environment);
//    System.out.println("DefaultInitialContext.DefaultInitialContext");
    if (myProps.get(Context.INITIAL_CONTEXT_FACTORY) != null) {
    // try to retrieve initial context factory from the environment
   _getDefaultInitCtxt();
  }
  }

  private Context _getDefaultInitCtxt() throws NamingException {
  if (defaultInitContext == null) {
    // from the environment try to find the right initialcontext provider.
    String initCtxtFactoryName = (String) myProps.get(Context.INITIAL_CONTEXT_FACTORY);
    if (initCtxtFactoryName == null) {
        // no scheme available + no factory set in the environment => will return our default implementation
        initCtxtFactoryName = "com.sap.engine.services.jndi.InitialContextFactoryImpl";
        myProps.put(Context.INITIAL_CONTEXT_FACTORY, initCtxtFactoryName);
      }
      // retrieve the initial context factory.
      InitialContextFactory fac = ((ProviderResolver) ProviderResolver.getInstance()).findInitialContextFactory(myProps);
      if (fac == null) {
        NamingException ne = new NoInitialContextException("failed to retrieve initial context factory " + initCtxtFactoryName);
        throw ne;
      }
      defaultInitContext = fac.getInitialContext(myProps);
    }
    return defaultInitContext;
  }

  // helper
  public static String _getURLScheme(String str) {
    int colon_posn = str.indexOf(':');
    int slash_posn = str.indexOf('/');

    if (colon_posn > 0 && (slash_posn == -1 || colon_posn < slash_posn)) {
      return str.substring(0, colon_posn);
    }

    return null;
  }

  // retrieve a URL context by scheme.
  private Context _getURLContext(String scheme, Hashtable env) throws NamingException {
    // optimize the search by keeping proxies in a cache!
    ObjectFactory fac = null;
    fac = _findURLContextFactory(scheme, env);

    if (fac == null) {
      return null; // which then means: default to JNDI standard behavior.
        //throw new NamingException("failed to find URL Context Factory for scheme "+scheme);
    }

    try {
      return (Context) fac.getObjectInstance(null, null, null, env);
    } catch (ClassCastException cce) {
      NamingException ne = new NamingException("URL context factory for scheme "+scheme+" did not return implementation of javax.naming.Context: "+cce);
      ne.setRootCause(cce);
      throw ne;
    } catch (Exception e) {
      NamingException ne = new NamingException("Error in URL context factory for scheme "+scheme);
      ne.setRootCause(e);
      throw ne;
    }
  }

  private ObjectFactory _findURLContextFactory(String scheme, Hashtable env) throws NamingException {
    return ((ProviderResolver) ProviderResolver.getInstance()).findURLContextFactory(scheme, env);
  }

  // end helper

  /**
   * We only worry about URL resolution, otherwise this InitialContext implementation
   * would not have been used. In that case, we look for an appropriate provider.
   */
  protected Context getURLOrDefaultInitCtx(String name) throws NamingException {
    String scheme = _getURLScheme(name);
    Context ctx = null;
    if (scheme != null) {
      ctx = _getURLContext(scheme, this.myProps);
      if (ctx != null) {
        return ctx;
      }
      // default logic in InitialContext
      ctx = NamingManager.getURLContext(scheme, myProps);
    if (ctx != null) {
        return ctx;
      }
     }
     // only if there is a scheme but we can not resolve it or if no scheme is provided
     return _getDefaultInitCtxt();
  }

  /**
   * Overwrites method in initial context implementation because we do not default too early
   * to the provider implementation (in getURLOrDefaultContxt of javax.naming.InitialContext);
   */
  protected Context getDefaultInitCtx() throws NamingException {
    return this;
  }

  /**
   * We only worry about URL resolution, otherwise this InitialContext implementation
   * would not have been used. In that case, we look for an appropriate resolver.
   */
  protected Context getURLOrDefaultInitCtx(Name name) throws NamingException {
    if (name.size() > 0) {
      String first = name.get(0);
      return getURLOrDefaultInitCtx(first);
    }
   //throw new NoInitialContextException();
    // no scheme available => will return our default implementation
    return _getDefaultInitCtxt();
  }
 // overwrites this methods because this is not the actual implemantation of InitialContext the user has requested
    public Object addToEnvironment(String propName, Object propVal) throws NamingException {
    myProps.put(propName, propVal);
    return _getDefaultInitCtxt().addToEnvironment(propName, propVal);
    }

    public Object removeFromEnvironment(String propName)
      throws NamingException {
  myProps.remove(propName);
  return _getDefaultInitCtxt().removeFromEnvironment(propName);
    }

    public Hashtable getEnvironment() throws NamingException {
  return _getDefaultInitCtxt().getEnvironment();
    }

    public void close() throws NamingException {
      myProps = null;
      _getDefaultInitCtxt().close();
      defaultInitCtx = null;
      gotDefault = false;
    }

    public String getNameInNamespace() throws NamingException {
    return _getDefaultInitCtxt().getNameInNamespace();
    }


}