package test.thesis.jndi.performance;

import java.util.Hashtable;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import test.thesis.jndi.provider.client.TestClient;
import test.thesis.jndi.provider.semantics.book.Book;
import test.thesis.jndi.provider.semantics.book.BookInfo;
import thesis.jndi.provider.api.DispatchStateFactory;
import thesis.jndi.provider.impl.ProviderResolver;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

public class PerformanceTest extends TestCase {
	
	public static final int NUMBER_OF_EXECUTIONS = 10000;
	
	private long[][] results = new long[2][4];
	
	private static final Properties bookInfo = new Properties();
	static {
		bookInfo.put(Book.AUTHOR_PROPERTY_NAME, "Joahn Rowling");
		bookInfo.put(Book.GENRE_PROPERTY_NAME, "adventure");
		bookInfo.put(Book.TITLE_PROPERTY_NAME, "Harry Potter and the deadly hallows");
		bookInfo.put(Book.YEAR_PROPERTY_NAME, "2007");
	}
	
	private static Hashtable env = new Hashtable();
	static {
		env.put(Context.INITIAL_CONTEXT_FACTORY, "test.thesis.jndi.provider.context.InitialInMemoryContextFactory");
		env.put(Context.URL_PKG_PREFIXES, "test.thesis.jndi.provider.context");
	}
	
	public PerformanceTest(String test) {
		super(test);
	}
	
	private void testInitializeProviderManager() throws NamingException  {
		 
	 System.out.println("Check if the provider is already registered");
	 assertFalse(javax.naming.spi.NamingManager.hasInitialContextFactoryBuilder());
	 
	 try {
	   ProviderResolver.init();
     assertTrue(javax.naming.spi.NamingManager.hasInitialContextFactoryBuilder());
     TestClient.registerProviders();
     System.out.println("Initialization was successful!");
   } catch (Exception e) {
     System.out.println("Error occured while initializing ProviderResolver. Exception is : " + e.toString());
     e.printStackTrace();
     throw new AssertionError("Cannot initialize ProviderManager.");
   }
   
  }
	
	private long testICAndUnbind() throws NamingException {
		System.gc();
		long time = System.currentTimeMillis();
		for (int i = 0; i < NUMBER_OF_EXECUTIONS; i++) {
		  (new InitialContext(env)).unbind("some_name");	
		}
		return System.currentTimeMillis() - time;
	}
	
	private long testICAndUnbindSchema() throws NamingException {
		System.gc();
		long time = System.currentTimeMillis();
		for (int i = 0; i < NUMBER_OF_EXECUTIONS; i++) {
		  (new InitialContext(env)).unbind("service:/some_name");
		}
				
		return System.currentTimeMillis() - time;
	}
	
	private long testBind(boolean toDispatch) throws NamingException {
		System.gc();
		Hashtable env1 = new Hashtable(env);
		if (toDispatch) {
		  env1.put(Context.STATE_FACTORIES, "thesis.jndi.provider.api.DispatchStateFactory");
		  env1.put(DispatchStateFactory.EXTRA_STATE_FACTORIES, "test.thesis.jndi.provider.semantics.book.BookStateFactory");
		} else {
			env1.put(Context.STATE_FACTORIES, "test.thesis.jndi.provider.semantics.book.BookStateFactory");
		}
		Book book = new Book(new BookInfo(bookInfo));
		
		Context ctx = new InitialContext(env1);
		long time = System.currentTimeMillis();
		for (int i = 0; i < NUMBER_OF_EXECUTIONS; i++) {
		  ctx.rebind("book", book);	
		}
		return System.currentTimeMillis() - time;
	}
	
	private long testLookup() throws NamingException {
		System.gc();
		Hashtable env1 = new Hashtable(env);
		env1.put(Context.OBJECT_FACTORIES, "test.thesis.jndi.provider.semantics.book.BookObjectFactory");
				
		Context ctx = new InitialContext(env1);
		long time = System.currentTimeMillis();
		for (int i = 0; i < NUMBER_OF_EXECUTIONS; i++) {
		  ctx.lookup("book");	
		}
		return System.currentTimeMillis() - time;
	}
	  
	public void testPreRun() {
		try {
		  System.out.println("Execute performance tests over standard JNDI framework");		
		  System.out.println("Check if the provider is already registered");
			assertFalse(javax.naming.spi.NamingManager.hasInitialContextFactoryBuilder());
		  
			long ic = testICAndUnbind();
			results[0][0] = ic;
			System.out.println("InitialContext + Unbind: " + NUMBER_OF_EXECUTIONS + " operations executed for " + ic + " millis");
			long ic_scheme = testICAndUnbindSchema();
			results[0][1] = ic_scheme;
			System.out.println("InitialContext + Unbind + Scheme: " + NUMBER_OF_EXECUTIONS + " operations executed for " + ic_scheme + " millis");
			long bind = testBind(false);
			results[0][2] = bind;
			System.out.println("Rebind: " + NUMBER_OF_EXECUTIONS + " operations executed for " + bind + " millis");
			long lookup = testLookup();
			results[0][3] = lookup;
			System.out.println("Lookup: " + NUMBER_OF_EXECUTIONS + " operations executed for " + lookup + " millis");
			
			System.out.println("Execute performance tests over JNDI Extension Module");
			testInitializeProviderManager();
			
			long ic2 = testICAndUnbind();
			results[1][0] = ic2;
			System.out.println("InitialContext + Unbind: " + NUMBER_OF_EXECUTIONS + " operations executed for " + ic2 + " millis");
			long ic_scheme2 = testICAndUnbindSchema();
			results[1][1] = ic_scheme2;
			System.out.println("InitialContext + Unbind + Scheme: " + NUMBER_OF_EXECUTIONS + " operations executed for " + ic_scheme2 + " millis");
			long bind2 = testBind(true);
			results[1][2] = bind2;
			System.out.println("Rebind: " + NUMBER_OF_EXECUTIONS + " operations executed for " + bind2 + " millis");
			long lookup2 = testLookup();
			results[1][3] = lookup2;
			System.out.println("Lookup: " + NUMBER_OF_EXECUTIONS + " operations executed for " + lookup2 + " millis");
					  
    } catch (Exception e) {
      e.printStackTrace();
      throw new AssertionError("Cannot invoke performanace test because of [" + e  + "]");
    }
  }
	
	
	public void testPerformance_InitialContext() {
		assertTrue("InitialContext + Unbind Performance JNDI Extension should be less or equal to standard JNDI ", (results[1][0] <= results[0][0]));
	}
	
	public void testPerformance_InitialContext_Schema() {
		assertTrue("InitialContext + Unbind + Scheme Performance JNDI Extension should be less or equal to standard JNDI ", (results[1][1] <= results[0][1]));
	}
	
	public void testPerformance_Bind() {
		assertTrue("Rebind Performance JNDI Extension should be less or equal to standard JNDI ", (results[1][2] <= results[0][2]));
	}
	
	public void testPerformance_Lookup() {
		assertTrue("Lookup Performance JNDI Extension should be less or equal to standard JNDI ", (results[1][3] <= results[0][3]));
	}
	
	
	
	public static Test suite(){	
		TestSuite suite = new TestSuite(); 
		suite.addTest(new PerformanceTest("testPreRun"));
		suite.addTest(new PerformanceTest("testPerformance_InitialContext"));
		suite.addTest(new PerformanceTest("testPerformance_InitialContext_Schema"));
		suite.addTest(new PerformanceTest("testPerformance_Bind"));
		suite.addTest(new PerformanceTest("testPerformance_Lookup"));
		return suite;
	}

}
