package test.thesis.jndi.provider.context;

import java.util.Hashtable;

import javax.naming.CannotProceedException;
import javax.naming.CompositeName;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.OperationNotSupportedException;
import javax.naming.spi.NamingManager;
import javax.naming.spi.ResolveResult;


public class ServiceSupportContext implements Context {
  protected Hashtable myEnv = null;
  
  protected ServiceSupportContext(Hashtable env) {
	// context that is not tied to any specific URL
	myEnv = env;  // copied on write
	//System.out.println("[ServiceSupportContext] was created with environment: " + env);
  }

  /**
   * For the "service" URL, the root URL is "service:/".
   */
  protected ResolveResult getRootURLContext(String url, Hashtable env) 
  throws NamingException {
	if (!url.startsWith("service:/")) {
	  throw new IllegalArgumentException(url + " is not a service URL");
	}

	String objName = url.length() > 9 ? url.substring(9) : null;

	// Represent object name as empty or single-component composite name.
	CompositeName remaining = new CompositeName();
	if (objName != null) {
	  remaining.add(objName);
	}

	// Get handle to the static namespace that we use for testing.
	// In an actual implementation, this might be the root
	// namespace on a particular server.
	
	Context ctx = InMemoryContext.getStaticNamespace(env);
    
	return (new ResolveResult(ctx, remaining));
  }

  /**
   * URL					Suffix
   * service://host:port				<empty string>
   * service://host:port/rest/of/name 		rest/of/name
   * service:///rest/of/name			rest/of/name
   * service:/rest/of/name			rest/of/name
   * service:rest/of/name			rest/of/name
   */
  protected Name getURLSuffix(String prefix, String url) throws NamingException {
	String suffix = url.substring(prefix.length());
	if (suffix.length() == 0) {
	  return new CompositeName();
	}

	if (suffix.charAt(0) == '/') {
	  suffix = suffix.substring(1); // skip leading slash
	}

	return new CompositeName().add(suffix);
  }

  /**
   * Finds the prefix of a URL.
   * URL					Prefix
   * service://host:port				service://host:port
   * service://host:port/rest/of/name 		service://host:port
   * service:///rest/of/name			service://
   * service:/rest/of/name			service:
   * service:rest/of/name			service:
   */
  protected String getURLPrefix(String url) throws NamingException {
	int start = url.indexOf(":");

	if (start < 0) {
	  throw new OperationNotSupportedException("Invalid URL: " + url);
	}
	++start; // skip ':'

	if (url.startsWith("//", start)) {
	  start += 2;  // skip double slash

	  // find last slash
	  int posn = url.indexOf("/", start);
	  if (posn >= 0) {
		start = posn;
	  } else {
		start = url.length();  // rest of URL
	  }
	}

	// else 0 or 1 initial slashes; start is unchanged
	return url.substring(0, start);
  }

  /**
   * Determines whether two URLs are the same.
   * Default implementation uses String.equals().
   * Subclass should override if this is not appropriate.
   * This method is used by rename(). 
   */
  protected boolean urlEquals(String url1, String url2) {
	return url1.equals(url2);
  }

  /**
   * Gets the context in which to continue the operation. This method
   * is called when this context is asked to process a multicomponent
   * Name in which the first component is a URL.
   * Treat the first component like a junction: resolve it and then use 
   * NamingManager.getContinuationContext() to get the target context in
   * which to operate on the remainder of the name (n.getSuffix(1)).
   */
  protected Context getContinuationContext(Name n) throws NamingException {
	Object obj = lookup(n.get(0));
	CannotProceedException cpe = new CannotProceedException();
	cpe.setResolvedObj(obj);
	cpe.setEnvironment(myEnv);
	return NamingManager.getContinuationContext(cpe);
  }

  public Object lookup(String name) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  return ctx.lookup(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public Object lookup(Name name) throws NamingException {
	if (name.size() == 1) {
	  return lookup(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		return ctx.lookup(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public void bind(String name, Object obj) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  ctx.bind(res.getRemainingName(), obj);
	} finally {
	  ctx.close();
	}
  }

  public void bind(Name name, Object obj) throws NamingException {
	if (name.size() == 1) {
	  bind(name.get(0), obj);
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		ctx.bind(name.getSuffix(1), obj);
	  } finally {
		ctx.close();
	  }
	}
  }

  public void rebind(String name, Object obj) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  ctx.rebind(res.getRemainingName(), obj);
	} finally {
	  ctx.close();
	}
  }

  public void rebind(Name name, Object obj) throws NamingException {
	if (name.size() == 1) {
	  rebind(name.get(0), obj);
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		ctx.rebind(name.getSuffix(1), obj);
	  } finally {
		ctx.close();
	  }
	}
  }

  public void unbind(String name) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  ctx.unbind(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public void unbind(Name name) throws NamingException {
	if (name.size() == 1) {
	  unbind(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		ctx.unbind(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public void rename(String oldName, String newName) throws NamingException {
	String oldPrefix = getURLPrefix(oldName);
	String newPrefix = getURLPrefix(newName);
	if (!urlEquals(oldPrefix, newPrefix)) {
	  throw new OperationNotSupportedException(
		  "Renaming using different URL prefixes not supported : " +
		  oldName + " " + newName);
	}

	ResolveResult res = getRootURLContext(oldName, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  ctx.rename(res.getRemainingName(), getURLSuffix(newPrefix, newName));
	} finally {
	  ctx.close();
	}
  }

  public void rename(Name name, Name newName) throws NamingException {
	if (name.size() == 1) {
	  if (newName.size() != 1) {
		throw new OperationNotSupportedException(
			"Renaming to a Name with more components not supported: " + newName);
	  }
	  rename(name.get(0), newName.get(0));
	} else {
	  // > 1 component with 1st one being URL
	  // URLs must be identical; cannot deal with diff URLs
	  if (!urlEquals(name.get(0), newName.get(0))) {
		throw new OperationNotSupportedException(
			"Renaming using different URLs as first components not supported: " +
			name + " " + newName);
	  }

	  Context ctx = getContinuationContext(name);
	  try {
		ctx.rename(name.getSuffix(1), newName.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public NamingEnumeration list(String name)	throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  return ctx.list(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public NamingEnumeration list(Name name) throws NamingException {
	if (name.size() == 1) {
	  return list(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		return ctx.list(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public NamingEnumeration listBindings(String name) 
  throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  return ctx.listBindings(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public NamingEnumeration listBindings(Name name) throws NamingException {
	if (name.size() == 1) {
	  return listBindings(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		return ctx.listBindings(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public void destroySubcontext(String name) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  ctx.destroySubcontext(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public void destroySubcontext(Name name) throws NamingException {
	if (name.size() == 1) {
	  destroySubcontext(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		ctx.destroySubcontext(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public Context createSubcontext(String name) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  return ctx.createSubcontext(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public Context createSubcontext(Name name) throws NamingException {
	if (name.size() == 1) {
	  return createSubcontext(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		return ctx.createSubcontext(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public Object lookupLink(String name) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  return ctx.lookupLink(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public Object lookupLink(Name name) throws NamingException {
	if (name.size() == 1) {
	  return lookupLink(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		return ctx.lookupLink(name.getSuffix(1));
	  } finally {
		ctx.close();
	  }
	}
  }

  public NameParser getNameParser(String name) throws NamingException {
	ResolveResult res = getRootURLContext(name, myEnv);
	Context ctx = (Context)res.getResolvedObj();
	try {
	  return ctx.getNameParser(res.getRemainingName());
	} finally {
	  ctx.close();
	}
  }

  public NameParser getNameParser(Name name) throws NamingException {
	if (name.size() == 1) {
	  return getNameParser(name.get(0));
	} else {
	  Context ctx = getContinuationContext(name);
	  try {
		return ctx.getNameParser(name.getSuffix(1));  
	  } finally {
		ctx.close();
	  }
	}
  }

  public String composeName(String name, String prefix)
  throws NamingException {
	if (prefix.equals("")) {
	  return name;
	} else if (name.equals("")) {
	  return prefix;
	} else {
	  return (prefix + "/" + name);
	}
  }

  public Name composeName(Name name, Name prefix) throws NamingException {
	Name result = (Name)prefix.clone();
	result.addAll(name);
	return result;
  }

  public String getNameInNamespace() throws NamingException {
	return ""; // A URL context's name is ""
  }

  public Object removeFromEnvironment(String propName)
  throws NamingException {
	if (myEnv == null) {
	  return null;
	}
	myEnv = (Hashtable)myEnv.clone();
	return myEnv.remove(propName);
  }

  public Object addToEnvironment(String propName, Object propVal)
  throws NamingException {
	myEnv = (myEnv == null) ? 	
		new Hashtable(11, 0.75f) : (Hashtable)myEnv.clone();
		return myEnv.put(propName, propVal);
  }

  public Hashtable getEnvironment() throws NamingException {
	if (myEnv == null) {
	  return new Hashtable(5, 0.75f);
	} else {
	  return (Hashtable)myEnv.clone();
	}
  }

  public void close() throws NamingException {
  }
}